@@ -12,31 +12,6 @@
 (*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *)
 (*  General Public License for more details.                              *)
 (**************************************************************************)
-	    
-let dic_lookup s p len = 
-  PersonalDict.lookup s p len @
-  Chasen.darts_lookup s p len
-
-let dic_lookup_prefix s p len =
-  PersonalDict.lookup_prefix s p len @ 
-    Chasen.darts_lookup_prefix s p len
-
-let connect_mrph ~state ~mrph =
-  let new_state, con_cost = 
-    Chasen.connect_cost state mrph in
-  if con_cost < 0 then raise Exit else
-  let hinsi_cost =
-    if Chasen.is_undef mrph = 0 then
-      let hinsi = Chasen.hinsi mrph in
-      Chasen.hinsi_cost hinsi
-    else
-      let undef_info = Chasen.undef_info mrph in
-      Chasen.undef_info_cost undef_info +
-	Chasen.undef_info_cost_step undef_info *
-	(Chasen.keyword_len mrph / 2) in
-  let mrph_cost = hinsi_cost * Chasen.weight mrph * Chasen.mrph_cost_weight in
-  let cost = mrph_cost + con_cost in
-  (new_state, cost)
 
 (********************)
 (* ܥդ *)
@@ -65,15 +40,20 @@ end
 module EdgeLabel = struct 
 
   (* cost = ǼΤΥ + ξ֤ؤϢܥ *)
-  type t = {mrph : Chasen.mrph; cost : int}
+  type t = {mrph : Mrph.mrph; pos : int; cost : int}
 
-  let default = {mrph = Chasen.null_mrph; cost = 0}
+  let make_terminal v =
+    {mrph = Mrph.null_mrph; cost = 0; pos = v.Vertex.pos}
+
+  let default = {mrph = Mrph.null_mrph; cost = 0; pos = -1}
 
   let compare v1 v2 =
+    let sgn = v1.pos - v2.pos in
+    if sgn <> 0 then sgn else
     let sgn = v1.cost - v2.cost in
     if sgn <> 0 then sgn else
     (* IDEA:mrphIDդ롩 *)
-    Chasen.compare_mrph v1.mrph v2.mrph
+    Mrph.compare_mrph v1.mrph v2.mrph
 
   let equal v1 v2 = (compare v1 v2 = 0)
 	
@@ -100,15 +80,14 @@ let parse string ~state ~pos ~len =
   for i = 0 to len - 1 do
     (* ǤΥꥹ *)
     let mrph_list = 
-      dic_lookup_prefix string (pos+i) (len - i) in
+      Mrph.lookup_prefix string (pos+i) (len - i) in
     (* ǽ̤θ *)
     let undef_list =
       (* ̤θʬΥѥեޥ󥹤ľ *)
       let r = ref [] in
       for j = 1  to len - i do
-
 	let u_list = 
-	  Chasen.undefwords (EucString.sub string (pos + i) j) in
+	  Mrph.undefwords (EucString.sub string (pos + i) j) in
 	r := List.rev_append u_list !r;
       done;
       !r in
@@ -117,8 +96,8 @@ let parse string ~state ~pos ~len =
     let add v (mrph, mlen) =
       try
 	let s', cost = 
-	  connect_mrph ~state:v.Vertex.state ~mrph in
-	let label = {EdgeLabel.mrph = mrph; EdgeLabel.cost = cost} in
+	  Mrph.connect_mrph ~state:v.Vertex.state ~mrph in
+	let label = {EdgeLabel.mrph = mrph; EdgeLabel.cost = cost; EdgeLabel.pos = v.Vertex.pos} in
 	let v' = create_vertex (pos+i+mlen) s' in
 	let e = TransGraph.E.create v label v' in
 	
@@ -142,7 +121,8 @@ let parse string ~state ~pos ~len =
   (* üĤ *)
   TransGraph.add_vertex graph Vertex.terminal;
   let add_termlink v =
-    let e = TransGraph.E.create v EdgeLabel.default Vertex.terminal in
+    let e = TransGraph.E.create v (EdgeLabel.make_terminal v)
+	Vertex.terminal in
     TransGraph.add_edge_e graph e in
   List.iter add_termlink vertex_array.(len);
   (* ֤Τϥդ *)
@@ -177,7 +157,7 @@ let convert_path pos path =
     | [] -> assert false in
   List.rev (f pos 0 [] path)  
 
-let best_path_from graph ~state ~pos =
+let best_path_from_parse_obj graph ~state ~pos =
     let fv = {Vertex.state = state; Vertex.pos = pos} in
     let path, cost = PathFinder.shortest_path graph fv Vertex.terminal in
     convert_path pos path
@@ -187,22 +167,22 @@ let best_path_from graph ~state ~pos =
 (************)
 
 let list_candidates string ~state ~pos ~mrph_len ~len =
-  let mrph_list = dic_lookup string pos mrph_len in
+  let mrph_list = Mrph.lookup string pos mrph_len in
   let mrph_list = 
-    Chasen.undefwords (EucString.sub string pos mrph_len) @ mrph_list in
+    Mrph.undefwords (EucString.sub string pos mrph_len) @ mrph_list in
   let f mrph =
-    let state', cost = connect_mrph ~state ~mrph in
+    let state', cost = Mrph.connect_mrph ~state ~mrph in
     (mrph, state', cost) in
   let list = List.map f mrph_list in
   let cmp (_, _, c1) (_, _, c2) = c1 - c2 in
   List.sort cmp list
 
-let list_candidates_from graph ~state ~pos ~mrph_len =
+let list_candidates_from_parse_obj graph ~state ~pos ~mrph_len =
   let v = {Vertex.state = state; Vertex.pos = pos} in
   let edge_list = TransGraph.succ_e graph v in
 
   let filter (_, label, _) = 
-    Chasen.keyword_len (label.EdgeLabel.mrph) = mrph_len in
+    EucString.length (Mrph.keyword (label.EdgeLabel.mrph)) = mrph_len in
 
   let compare (_, label1, _) (_, label2, _) = 
     label1.EdgeLabel.cost - label2.EdgeLabel.cost in
@@ -216,3 +196,63 @@ let list_candidates_from graph ~state ~pos ~mrph_len =
     (List.sort compare
        (* keyword_lenmrph_len˰פΤӤ*)
        (List.filter filter edge_list))
+
+(********)
+(* ؽ *)
+(********)
+
+(* Ǥ *)
+let remove_last list =
+  match List.rev list with
+    [] -> []
+  | _ :: list -> List.rev list
+
+let learn graph ~state ~pos path =
+
+  (* pathˤǰʳϥ̵Ȥ롣*)
+  let module Cost = struct
+    type label = EdgeLabel.t
+    type t = Int of int | Inf
+
+    let weight label = 
+      let is_terminal = 
+	Mrph.compare_mrph label.EdgeLabel.mrph Mrph.null_mrph = 0 in
+      if is_terminal then Int 0 else
+      let pos = label.EdgeLabel.pos in
+      let s_form = Mrph.surface_form label.EdgeLabel.mrph in
+      let length = EucString.length (Mrph.keyword (label.EdgeLabel.mrph)) in
+      let match_fun (s, p, len) =
+	p = pos && len = length && s = s_form in
+      if List.exists match_fun path then
+	Int label.EdgeLabel.cost
+      else
+	Inf
+
+    let zero = Int 0
+
+    let add c1 c2 = 
+      match c1, c2 with
+	Int c1, Int c2 -> Int (c1 + c2)
+      | Inf, _ | _, Inf  -> Inf
+
+    let compare c1 c2 =
+      match c1, c2 with
+	Int c1, Int c2 -> c1 - c2
+      | Inf, Inf -> 0
+      | Inf, _ -> 1
+      | _, Inf -> -1
+  end in
+  
+  let module PathFinder = Graph.Path.Dijkstra(TransGraph)(Cost) in
+
+  let fv = {Vertex.state = state; Vertex.pos = pos} in
+
+  let path, _ =
+    try 
+      PathFinder.shortest_path graph fv Vertex.terminal 
+    with Not_found -> [], Cost.Inf in
+  let f e = 
+    let mrph = (TransGraph.E.label e).EdgeLabel.mrph in
+(*    Printf.eprintf "%s\n" (Mrph.surface_form mrph); flush stderr;*)
+    Mrph.incr_count mrph in
+  List.iter f (remove_last path)
