@@ -4,6 +4,7 @@ open ScmTypes
 exception UnknownCmd of string
 exception Quit
 exception SyntaxError
+exception ScmError of string
 
 type eval_method = Lazy_eval | Eager_eval
 
@@ -11,6 +12,7 @@ let current_input_port = ref (datum_of_in_channel stdin)
 let current_output_port = ref (datum_of_out_channel stdout)
 
 let print_datum d = print_endline (string_of_datum d)
+let verbose_eval = ref false
 
 module Environment =
 struct
@@ -46,19 +48,6 @@ struct
       
   let get env k = 
     let vr = lookup env k in !vr
-      
-  (* let rec bind f =
-    function
-      Nil,Nil -> ()
-      | Pair p1,Pair p2 ->
-	  f (string_of_symbol p1.car) p2.car;
-	  if is_symbol p1.cdr then
-	    f (string_of_symbol p1.cdr) p2.cdr
-	  else
-	    bind f (p1.cdr, p2.cdr)
-      | Symbol x, r -> f x r
-      | a,b -> invalid_arg 
-	  ("cannot bind " ^ (string_of_datum b) ^ " on " ^ (string_of_datum a)) *)
 
   let rec bind signature params prefs =
     let rec loop =
@@ -66,10 +55,14 @@ struct
 	| Pair sp,Pair pp,hd::tl ->
 	    hd := pp.car;
 	    loop (sp.cdr,pp.cdr,tl)
-	| Symbol _, v, hd::[] ->
-	    hd := v
+	| Symbol _, v, hd::rest ->
+	    hd := v;
+	    List.iter (fun v -> v := Nil) rest;
 	| Nil,Nil,[] -> ()
-	| _ -> failwith ("bind, sig:" ^ string_of_datum signature)
+	| s,p,l -> 
+	    print_endline (string_of_datum s);
+	    print_endline (string_of_datum p);
+	    failwith ("bind, sig:" ^ string_of_datum signature ^ ",params: "  ^ string_of_datum params )
     in loop (signature,params,prefs)
 
   let add env k v = 
@@ -221,7 +214,7 @@ let rec eval (env : environment) (cont : datum -> datum) =
 		 f env cont c.cdr 
 	     | datum -> failwith ("expected procedure, got " ^ string_of_datum datum))
 	in 
-	  (* print_datum l; *)
+	  if !verbose_eval then print_datum l; 
 	  eval env (fun v -> evaluate v) (c.car)
     | value -> cont value
 and scm_begin env cont  =
@@ -258,7 +251,7 @@ let scm_cond env cont args =
 		  (* | Nil -> cont Nil *)
 		  | Symbol "=>" ->
 		      eval env cont (cons (caddr c) test_result)
-		  | _ -> eval env cont (cadr c)
+		  | _ -> scm_begin env cont (cdr c)
 	    else
 	      loop cond_clausel.cdr 
 	  in
@@ -293,7 +286,7 @@ let rec scm_define env cont args =
       let fun_name = string_of_symbol scm_fun_name in
 	let lambda = scm_lambda env (fun x -> x) (cons signature value) in
 	let _ = Environment.global_add env fun_name lambda in
-	  print_endline fun_name;
+	  (* print_endline fun_name; *)
 	  cont scm_fun_name;
     else
       let var_name = string_of_symbol key in
@@ -413,6 +406,7 @@ let scm_do env cont args =
     function 
       | Nil -> 
 	  let refs = List.map (fun v -> ref v) values in
+	  let rev_refs = List.rev refs in
 	  let env1 = Environment.add_ref_binding env keys refs in
 	  let test = car test_line in
 	  let rec iterate v =
@@ -420,17 +414,20 @@ let scm_do env cont args =
 	    else
 	      scm_begin env1
 		(fun _ ->
-		   let rec incr_vars = 
+		   let rec incr_vars acc = 
 		     function
-		       | r::rtl,step::steptl ->
-			   if step  = Nil then incr_vars (rtl,steptl)
+		       | step::steptl ->
+			   if step  = Nil then incr_vars (None::acc) steptl
 			   else eval env1 
-			     (fun v -> 
-				r := v;
-				incr_vars (rtl,steptl))
+			     (fun v -> incr_vars (Some v::acc) steptl)
 			     step
-		       | _ -> eval env1 iterate test 
-		   in incr_vars (refs,steps)) body
+		       | [] -> 
+			   List.iter2 (fun r vopt -> 
+					 match vopt with
+					     Some v -> r := v
+					   | _ -> ()) rev_refs acc ;
+			   eval env1 iterate test 
+		   in incr_vars [] steps) body
 	  in eval env1 iterate test
       | Pair p ->
 	  let init_line = p.car in
@@ -886,6 +883,12 @@ let scm_getenv =
        try String (Sys.getenv (string_of_scm_string s))
        with Not_found -> scm_false)
 
+let scm_file_exists =
+  simple_cmd_1 (fun s -> datum_of_bool (Sys.file_exists (string_of_scm_string s)))
+
+let scm_chdir =
+  simple_cmd_1 (fun s -> Sys.chdir (string_of_scm_string s); (String (Sys.getcwd ())))
+
 let scm_vector_ref args =
   let v = car args
   and k = cadr args in
@@ -1012,8 +1015,10 @@ let output_channel_of_args args =
 let scm_read args =
   let ich = input_channel_of_args args in
   let lexbuf = Lexing.from_channel ich in
-    try ScmParser.item ScmLexer.token lexbuf
-    with End_of_file -> Eof
+    try let result = ScmParser.item ScmLexer.token lexbuf in
+      quote (clamp result)
+    with Exit -> Eof
+
 
 let scm_read_char args =
   let ich = input_channel_of_args args in
@@ -1073,6 +1078,7 @@ let load env ?(prompt = fun v -> v) stream =
   let result = ref Unspecified in
   let rec rep () =
     let read_block = ScmParser.item ScmLexer.token stream in
+    (* let identity x = x in *)
       result := eval env prompt read_block;
       rep ()
   in 
@@ -1084,7 +1090,7 @@ let scm_load env cont args =
   let do_it scm_string =
     let file_name = string_of_scm_string scm_string in
     let stream = Lexing.from_channel (open_in file_name) in
-      cont (load env ~prompt:cont stream)
+      cont (load env ~prompt:(fun x -> x) stream)
   in eval env do_it (car args)
 
 
@@ -1096,6 +1102,10 @@ let define_scm_var env name value =
 let register_scm_fun env name func =
   Environment.add env name (ML_EagerProcedure func)  
 
+let scm_debug _ = verbose_eval := true; Unspecified
+let scm_error = 
+  simple_cmd_1 (fun s -> raise (ScmError (string_of_datum s)))
+
 let rec command_list = 
   ["+", ML_EagerProcedure scm_add;
    "-", ML_EagerProcedure scm_sub;
@@ -1118,7 +1128,7 @@ let rec command_list =
    "symbol?", ML_EagerProcedure scm_symbol_p;
    "boolean?", ML_EagerProcedure scm_boolean_p;
    "number?", ML_EagerProcedure scm_number_p;
-   "vector?", ML_EagerProcedure scm_vector_p;
+   "vector", ML_EagerProcedure scm_vector_p;
    "pair?", ML_EagerProcedure scm_pair_p;
    "string?", ML_EagerProcedure scm_string_p;
    "port?", ML_EagerProcedure scm_port_p;
@@ -1300,6 +1310,10 @@ let rec command_list =
    "write-char", ML_EagerProcedure scm_write_char;
    "scheme-report-environment" , ML_EagerProcedure scm_report_environment;
    "truncate", ML_EagerProcedure scm_truncate;
+   "file-exists?", ML_EagerProcedure scm_file_exists;
+   "_debug", ML_EagerProcedure scm_debug;
+   "error", ML_EagerProcedure scm_error;
+   "chdir", ML_EagerProcedure scm_chdir;
   ]  
 and scm_eval env cont args =
   let eval_code code =
